from __future__ import division
from __future__ import print_function

import glob
import time
import random
import argparse
import numpy as np
import torch
import torch.nn.functional as F
import torch.optim as optim
from sklearn.metrics import accuracy_score
from torch.autograd import Variable
import scipy.sparse as sp
import matplotlib.pyplot as plt

from utils_old import load_data,  normalize_adj, findNearLabelNodes,\
    computeCluterCenter, constrainCov, computeEntropy, accuracy, stast, load_data2

from models import GAT, SpGAT

import os
os.environ['CUDA_LAUNCH_BLOCKING'] = "1"
'''
本代码是在GAT基础上修改而来的。新工作idea：首先学习不用的映射，其对应不同的聚类。借助attention机制计算得到节点的embedding。在计算
loss部分，分为两块：传统的交叉熵loss，选择性loss，即对于不同的无标签节点，计算其在不同映射下embedding与类中心的距离并计算熵。我们希望
节点的熵越大越好，这样可以挑选出节点属于某个类。

具体步骤：
1.按照GAT，第一层设置head= #c,  即#c个映射（W）。
2.第二层设置head=1, 即一个映射W。
3.得到的embedding，带label的点，计算交叉熵，无labels的点计算与聚类中心的熵。
'''

# Training settings
parser = argparse.ArgumentParser()
parser.add_argument('--no-cuda', action='store_true', default=False, help='Disables CUDA training.')
parser.add_argument('--fastmode', action='store_true', default=False, help='Validate during training pass.')
parser.add_argument('--isSample', action='store_true', default=True, help='running sample dataset.')
parser.add_argument('--rate', type=float, default=0.005, help='sample rate for training model')
parser.add_argument('--valRate', type=float, default=0.3, help='samples for validation.')

parser.add_argument('--sparse', action='store_true', default=True, help='GAT with sparse version or not.')
parser.add_argument('--seed', type=int, default=2, help='Random seed.')
parser.add_argument('--epochs', type=int, default=200, help='Number of epochs to train.')
parser.add_argument('--lr', type=float, default=0.002, help='Initial learning rate.')
parser.add_argument('--weight_decay', type=float, default=1e-4, help='Weight decay (L2 loss on parameters).')
parser.add_argument('--hidden', type=int, default=[32, 16, 8],nargs=3, help='Number of hidden units.') # each dimensions [64, 32, 16]
# parser.add_argument('--nb_heads', type=int, default=7, help='the number of clusters.')
parser.add_argument('--dropout', type=float, default=0.5, help='Dropout rate (1 - keep probability).')
parser.add_argument('--alpha', type=float, default=0.2, help='Alpha for the leaky_relu.')
parser.add_argument('--patience', type=int, default=100, help='Patience')
parser.add_argument('--outDim', type=int, default=16, help='final dim of output, (hidden[0] + hidden[1])==>outDim')
parser.add_argument('--k_nn', type=int, default=5, help='select k nn from each community for each node without label')
parser.add_argument('--dataset',type=str, default='pubmed',help='dataset')
parser.add_argument('--hpara_alpha',type=float, default=0.1,help='cross entropy')
parser.add_argument('--hpara_beta',type=float,default=2,help='entropy')
parser.add_argument('--hpara_row',type=float,default=0.001,help='parameters l2 loss')
parser.add_argument('--hpara_gamma',type=float,default=1,help='parameters cov')
args = parser.parse_args()
dataset = args.dataset

alpha = args.hpara_alpha # cross entropy 1
beta = args.hpara_beta # entropy 2
row = args.hpara_row # parameters l2 loss  这项不重要。
gamma = args.hpara_gamma # parameters cov

args.cuda = not args.no_cuda and torch.cuda.is_available()
args.cuda = False

CUDA_LAUNCH_BLOCKING=1
random.seed(args.seed)
np.random.seed(args.seed)
torch.manual_seed(args.seed)
if args.cuda:
    torch.cuda.manual_seed(args.seed)

# Load data
if args.isSample is False:
    # adj, features, labels, idx_train, idx_val, idx_test = load_data()
    # labels: N; idx_train: len(training); idx_val: len(validation); idx_test: len(test)
    adj, features, y_train, y_val, y_test, train_mask, val_mask, test_mask = load_data(dataset)
    # features = normalize_features(features.todense()) ## normalize features.
    # features = torch.from_numpy(features)
    features = torch.Tensor(features.todense())
    adj = normalize_adj(adj + sp.eye(adj.shape[0])).todense()

    idx_train = np.where(train_mask == True)[0]
    idx_val = np.where(val_mask == True)[0]
    idx_test = np.where(test_mask == True)[0]

    train_labels = torch.max(torch.LongTensor(y_train[idx_train]), 1)[1]
    val_labels = torch.max(torch.LongTensor(y_val[idx_val]), 1)[1]
    test_labels = torch.max(torch.LongTensor(y_test[idx_test]), 1)[1]
    labels = torch.max(torch.LongTensor(y_test + y_val + y_train), 1)[1]


    # Model and optimizer
    if args.sparse:
        model = SpGAT(nfeat=features.shape[1],
                    nhid=args.hidden,
                    outDim=args.outDim,
                    nclass=int(labels.max()) + 1,
                    dropout=args.dropout,
                    alpha=args.alpha)
    else:
        model = GAT(nfeat=features.shape[1],
                    nhid=args.hidden,
                    outDim=args.outDim,
                    nclass=int(labels.max()) + 1,
                    dropout=args.dropout,
                    alpha=args.alpha)
    optimizer = optim.Adam(model.parameters(),
                           lr=args.lr,
                           weight_decay=args.weight_decay)

    adj = torch.Tensor(adj)
    idx_train = idx_train.astype(np.int64)
    idx_val = idx_val.astype(np.int64)
    idx_test = idx_test.astype(np.int64)
    idx_train = torch.tensor(idx_val)
    idx_test = torch.tensor(idx_val)
    idx_val = torch.tensor(idx_val)
    y_train = torch.Tensor(y_train)
    y_val = torch.Tensor(y_val)
    y_test = torch.Tensor(y_test)
    if args.cuda:
        model.cuda()
        features = features.cuda()
        adj = adj.cuda()

        idx_train = idx_train.cuda()
        idx_val = idx_val.cuda()
        idx_test = idx_test.cuda()
        labels = labels.cuda()
        y_train = y_train.cuda()
        y_val = y_val.cuda()
        y_test = y_test.cuda()
        train_labels = train_labels.cuda()
        val_labels = val_labels.cuda()
        test_labels = test_labels.cuda()

else:
    adj, features, idx_train, idx_test, labels, minclusterNum = load_data2(dataset=dataset, rate=args.rate, k=args.k_nn)
    # adj, features, idx_train, idx_test, labels, minclusterNum = load_data_random(dataset=dataset, rate=args.rate)
    adj_, features_, idx_val, idx_test_, labels_, minclusterNumVal = load_data2(dataset=dataset, rate=args.valRate, k=args.k_nn)
    # adj, features, y_train, y_val, y_test, train_mask, val_mask, test_mask = load_data(dataset)
    # features = normalize_features(features.todense()) ## normalize features.
    # features = torch.from_numpy(features)
    features = torch.Tensor(features.todense())
    adj = normalize_adj(adj + sp.eye(adj.shape[0])).todense()

    # labels = torch.Tensor(labels)

    minclusterNum = int(minclusterNum if minclusterNum < args.k_nn else args.k_nn)
    minclusterNumVal = int(minclusterNumVal if minclusterNumVal < args.k_nn else args.k_nn)

    # Model and optimizer
    if args.sparse:
        model = SpGAT(nfeat=features.shape[1],
                      nhid=args.hidden,
                      outDim=args.outDim,
                      nclass=int(labels.max()) + 1,
                      dropout=args.dropout,
                      alpha=args.alpha)

    optimizer = optim.Adam(model.parameters(),
                           lr=args.lr,
                           weight_decay=args.weight_decay)
    adj = torch.Tensor(adj)
    idx_train = idx_train.to(torch.long)
    idx_test = idx_test.to(torch.long)
    idx_val = idx_val.to(torch.long)

    if args.cuda:
        model.cuda()
        features = features.cuda()
        adj = adj.cuda()
        idx_train = idx_train.to(torch.long).cuda()
        idx_test = idx_test.to(torch.long).cuda()
        labels = labels.cuda()
        idx_val = idx_val.to(torch.long).cuda()

nclass = torch.unique(labels).size()[0]
features, adj, labels = Variable(features), Variable(adj), Variable(labels)

def train(epoch):
    t = time.time()
    model.train()
    optimizer.zero_grad()
    output, H_middle = model(features, adj) # N x (dim x class):[ d1_1, d1_2, ...d1_n, | d2_1, d2_2, ..d2_n,| ...dc_1, dc_2, .. dc_n] for each row

    clusterCenter_train = computeCluterCenter(output, idx_train, labels[idx_train])  # nclass x out

    loss_cov_train = constrainCov(H_middle, idx_train, labels[idx_train], nclass)

    loss_entropy_train = computeEntropy(clusterCenter_train, output, idx_train)

    # loss_param_l2 = torch.Tensor([0.0]).cuda()
    loss_param_l2 = torch.Tensor([0.0])
    for name, param in model.named_parameters():
        if 'attention' in name:
            loss_param_l2 += torch.sum(param ** 2) / 2

    # output = F.log_softmax(output, dim=2)
    output_train = output[labels[idx_train], idx_train,:] # N_train x dim
    loss_cross = F.nll_loss(output_train, labels[idx_train])
    loss_train = alpha * loss_cross + beta * loss_entropy_train + row * loss_param_l2[0] + gamma * loss_cov_train

    acc_train = accuracy(output_train, labels[idx_train])
    loss_train.backward()
    optimizer.step()

    if not args.fastmode:
        # Evaluate validation set performance separately,
        # deactivates dropout during validation run.
        model.eval()
        output, H_middle = model(features, adj)

    # ## l2 normalization

    # output = F.log_softmax(output,dim=2) ## final dimension
    nearLabelNodes, nearLabelDist = findNearLabelNodes(output, idx_train, labels[idx_train], k=minclusterNum, restIdx=idx_val)
    nearLabelDist = nearLabelDist.permute([1, 0, 2]).reshape(
        (nearLabelDist.shape[1], nearLabelDist.shape[0] * nearLabelDist.shape[2]))  # N(val) x (nclass x dim)
    # for each node, each k elements is cluster.
    nearNodeDist = torch.topk(nearLabelDist, k=minclusterNum, dim=1, largest=False)  # the smallest nodes
    # nearNodeDist[0]: is the k smallest distance; nearNodeDist[1]: their location.

    # nearLabels = nearNodeDist[1] / args.k_nn  # N(unlabel) x k_nn (label)
    prelabels_val = stast(nearNodeDist, minclusterNum)
    # prelabels_val = stast(nearLabels.detach().cpu().numpy(), nearLabelDist[0])

    # clusterCenter_train = comput eCluterCenter(output, idx_train, train_labels) # nclass x out
    # prelabels_val = nearestEmbedding(output[:, idx_val,:], clusterCenter_train)
    # acc_val = accuracy_score(prelabels_val.detach().cpu().numpy(), labels[idx_val].detach().cpu().numpy())

    acc_val = accuracy_score(prelabels_val, labels[idx_val].detach().cpu().numpy())
    print('Epoch: {:04d}'.format(epoch+1),
          'loss_train: {:.4f}'.format(loss_train.data),
          '=loss_entropy_train: {:.4f}'.format(loss_entropy_train),
          '+loss_cross_entropy: {:.4f}'.format(loss_cross),
          '+loss_param_l2: {:.4f}'.format(loss_param_l2[0]),
          '+loss_cov_train: {:.4f}'.format(loss_cov_train),
          'acc_train: {:.4f}'.format(acc_train),)
    print('\t acc_val: {:.4f}'.format(acc_val),'time: {:.4f}s'.format(time.time() - t))

    # return loss_train.data
    return -acc_val

def compute_test():
    model.eval()
    output, H_middle = model(features, adj)

    nearLabelNodes, nearLabelDist = findNearLabelNodes(output, idx_train, labels[idx_train], k=minclusterNum,
                                                       restIdx=idx_test)
    nearLabelDist = nearLabelDist.permute([1, 0, 2]).reshape(
        (nearLabelDist.shape[1], nearLabelDist.shape[0] * nearLabelDist.shape[2]))  # N(unlabel) x (nclass x dim)
    nearNodeDist = torch.topk(nearLabelDist, k=minclusterNum, dim=1, largest=False)  # the smallest nodes
    prelabels_test = stast(nearNodeDist, minclusterNum)

    # clusterCenter_train = computeCluterCenter(output, idx_train, train_labels)
    # prelabels_test = nearestEmbedding(output[:, idx_test, :], clusterCenter_train)
    # acc_test = accuracy_score(prelabels_test.detach().cpu().numpy(), labels[idx_test].detach().cpu().numpy())
    acc_test = accuracy_score(prelabels_test, labels[idx_test].detach().cpu().numpy())
    print("Test set results:",  "accuracy= {:.4f}".format(acc_test))
          # "loss= {:.4f}".format(loss_test.data),

    # fig, ax = plt.subplots(3, 1)
    # H = H_middle
    # H_all = torch.zeros_like(H)
    # H_sub = H[idx_train, :]
    # H_ = torch.zeros_like(H_sub)
    # label_train = labels[idx_train]
    # ind = 0
    # hidx = 0
    #
    # for clusterId in range(len(torch.unique(label_train))):
    #     H_[hidx : (hidx + int(torch.sum(labels[idx_train] == clusterId))), :] = H_sub[label_train == clusterId, :]
    #     H_all[ind: (ind + torch.sum(labels == clusterId)), :] = H[labels == clusterId, :]  # N x dim
    #     ind += torch.sum(labels == clusterId)
    #     hidx += int(torch.sum(labels[idx_train] == clusterId))
    #
    # sim = H_ @ H_.t()
    # c = ax[0].pcolor(sim.detach().cpu().numpy(), cmap='RdBu')
    # ax[0].set_title('sim')
    #
    # z = ax[1].pcolor(H_.detach().cpu().numpy(), cmap='RdBu')
    # ax[1].set_title('labelled node embedding')
    #
    # H_all = H_all.t()
    # s = ax[2].pcolor(H_all.detach().cpu().numpy(), cmap='RdBu')
    # ax[2].set_title('all embedding')
    # fig.colorbar(s)
    #
    # fig.tight_layout()
    # plt.show()
    # figname = 'epochs' + str(args.epochs) + 'lr' + str(args.lr) + 'wdecay' + str(args.weight_decay) \
    #           + 'nhid' + str(args.hidden[0]) + '_' + str(args.hidden[1]) + '_' + str(args.hidden[2]) \
    #           + 'k_nn' + str(args.k_nn) + 'alpha' + str(alpha) + 'beta' + str(beta) + 'row' + str(row)\
    #           + 'gamma' + str(gamma) + '.jpg'
    # plt.savefig(figname)


# Train model
t_total = time.time()
loss_values = []
bad_counter = 0
best = args.epochs + 1
best_epoch = 0

for epoch in range(args.epochs):
    loss_values.append(train(epoch))

    torch.save(model.state_dict(), '{}.pkl'.format(epoch))
    if loss_values[-1] < best:
        best = loss_values[-1]
        best_epoch = epoch
        bad_counter = 0
    else:
        bad_counter += 1

    if bad_counter == args.patience:
        break

    files = glob.glob('*.pkl')
    for file in files:
        epoch_nb = int(file.split('.')[0])
        if epoch_nb < best_epoch:
            os.remove(file)

files = glob.glob('*.pkl')
for file in files:
    epoch_nb = int(file.split('.')[0])
    if epoch_nb > best_epoch:
        os.remove(file)

print("Optimization Finished!")
print("Total time elapsed: {:.4f}s".format(time.time() - t_total))

# Restore best model
print('Loading {}th epoch'.format(best_epoch))
model.load_state_dict(torch.load('{}.pkl'.format(best_epoch)))

# Testing
compute_test()

os.system('rm *.pkl')
