import numpy as np
import scipy.sparse as sp
import torch
import random
import itertools as itertls
import torch.nn.functional as F
import pickle as pkl
import networkx as nx
import random
import sys
from numpy import amax

def encode_onehot(labels):
    classes = set(labels)
    classes_dict = {c: np.identity(len(classes))[i, :] for i, c in enumerate(classes)}
    labels_onehot = np.array(list(map(classes_dict.get, labels)), dtype=np.int32)
    return labels_onehot

def sample_mask(idx, l):
    """Create mask."""
    mask = np.zeros(l)
    mask[idx] = 1
    return np.array(mask, dtype=np.bool)

# def load_data(path="./data/cora/", dataset="cora"):
#     """Load citation network dataset (cora only for now)"""
#     print('Loading {} dataset...'.format(dataset))
#
#     idx_features_labels = np.genfromtxt("{}{}.content".format(path, dataset), dtype=np.dtype(str))
#     features = sp.csr_matrix(idx_features_labels[:, 1:-1], dtype=np.float32)
#     labels = encode_onehot(idx_features_labels[:, -1])
#
#     # build graph
#     idx = np.array(idx_features_labels[:, 0], dtype=np.int32)
#     idx_map = {j: i for i, j in enumerate(idx)}
#     edges_unordered = np.genfromtxt("{}{}.cites".format(path, dataset), dtype=np.int32)
#     edges = np.array(list(map(idx_map.get, edges_unordered.flatten())), dtype=np.int32).reshape(edges_unordered.shape)
#     adj = sp.coo_matrix((np.ones(edges.shape[0]), (edges[:, 0], edges[:, 1])), shape=(labels.shape[0], labels.shape[0]), dtype=np.float32)
#
#     # build symmetric adjacency matrix
#     adj = adj + adj.T.multiply(adj.T > adj) - adj.multiply(adj.T > adj)
#
#     features = normalize_features(features)
#     adj = normalize_adj(adj + sp.eye(adj.shape[0]))
#
#     idx_train = range(140)
#     idx_val = range(200, 500)
#     idx_test = range(500, 1500)
#
#     adj = torch.FloatTensor(np.array(adj.todense()))
#     features = torch.FloatTensor(np.array(features.todense()))
#     labels = torch.LongTensor(np.where(labels)[1])
#
#     idx_train = torch.LongTensor(idx_train)
#     idx_val = torch.LongTensor(idx_val)
#     idx_test = torch.LongTensor(idx_test)
#
#     return adj, features, labels, idx_train, idx_val, idx_test

def parse_index_file(filename):
    """Parse index file."""
    index = []
    for line in open(filename):
        index.append(int(line.strip()))
    return index

def sample_mask(idx, l):
    """Create mask."""
    mask = np.zeros(l)
    mask[idx] = 1
    return np.array(mask, dtype=np.bool)

def load_data(dataset_str):  # {'pubmed', 'citeseer', 'cora'}
    """Load data."""
    names = ['x', 'y', 'tx', 'ty', 'allx', 'ally', 'graph']
    objects = []
    for i in range(len(names)):
        with open("data/ind.{}.{}".format(dataset_str, names[i]), 'rb') as f:
            if sys.version_info > (3, 0):
                objects.append(pkl.load(f, encoding='latin1'))
            else:
                objects.append(pkl.load(f))

    x, y, tx, ty, allx, ally, graph = tuple(objects)
    test_idx_reorder = parse_index_file("data/ind.{}.test.index".format(dataset_str)) # test ids
    test_idx_range = np.sort(test_idx_reorder) # change location of test into sort.

    if dataset_str == 'citeseer':
        # Fix citeseer dataset (there are some isolated nodes in the graph)
        # Find isolated nodes, add them as zero-vecs into the right position
        test_idx_range_full = range(min(test_idx_reorder), max(test_idx_reorder) + 1)
        tx_extended = sp.lil_matrix((len(test_idx_range_full), x.shape[1]))
        tx_extended[test_idx_range - min(test_idx_range), :] = tx
        tx = tx_extended
        ty_extended = np.zeros((len(test_idx_range_full), y.shape[1]))
        ty_extended[test_idx_range - min(test_idx_range), :] = ty
        ty = ty_extended

    features = sp.vstack((allx, tx)).tolil() ## all features
    features[test_idx_reorder, :] = features[test_idx_range, :]
    adj = nx.adjacency_matrix(nx.from_dict_of_lists(graph))


    labels = np.vstack((ally, ty))
    labels[test_idx_reorder, :] = labels[test_idx_range, :]

    idx_test = test_idx_range.tolist()
    idx_train = range(len(y))
    idx_val = range(len(y), len(y) + 500)

    train_mask = sample_mask(idx_train, labels.shape[0])
    val_mask = sample_mask(idx_val, labels.shape[0])
    test_mask = sample_mask(idx_test, labels.shape[0])

    y_train = np.zeros(labels.shape)
    y_val = np.zeros(labels.shape)
    y_test = np.zeros(labels.shape)
    y_train[train_mask, :] = labels[train_mask, :]
    y_val[val_mask, :] = labels[val_mask, :]
    y_test[test_mask, :] = labels[test_mask, :]

    print(adj.shape)
    print(features.shape)

    return adj, features, y_train, y_val, y_test, train_mask, val_mask, test_mask

def normalize_adj(mx):
    """Row-normalize sparse matrix"""
    rowsum = np.array(mx.sum(1))
    r_inv_sqrt = np.power(rowsum, -0.5).flatten()
    r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
    r_mat_inv_sqrt = sp.diags(r_inv_sqrt)
    return mx.dot(r_mat_inv_sqrt).transpose().dot(r_mat_inv_sqrt)


def normalize_features(mx):
    """Row-normalize sparse matrix"""
    rowsum = np.array(mx.sum(1))
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    mx = r_mat_inv.dot(mx)
    return mx

def normalize_features_gpu(mx, dim):
    """Row-normalize sparse matrix, if dim=1, else is column-normalize"""
    rowsum = torch.sum(mx, dim=dim)
    ## 这里需要一个normalization on gpu
    r_inv = rowsum**(-1)
    r_inv = r_inv.flatten()
    r_inv[torch.isinf(r_inv)] = 0
    r_mat_inv = torch.diag(r_inv)
    mx2 = r_mat_inv@mx
    return mx2

def normalize_feature_maxminGPU(mx):
    """Row-max normalization"""

    rowmax = torch.max(mx,dim=1).values
    rowmin = torch.min(mx,dim=1).values

    max = rowmax.unsqueeze(0).repeat(mx.shape[1], 1).t()
    min = rowmin.unsqueeze(0).repeat(mx.shape[1], 1).t()

    normal_max_min = (mx - min)/(max-min + 1e-8)

    return  normal_max_min

def normalize_feature_zscore(mx):
    rowmean = torch.mean(mx, dim=1)
    rowcov = torch.std(mx, dim=1)

    rowmean = rowmean.unsqueeze(1).repeat(1, mx.shape[1])
    rowstd = rowcov.unsqueeze(1).repeat(1, mx.shape[1])
    normal = (mx - rowmean)/(rowstd + 1e-8)
    return  normal

def accuracy(output, labels):
    preds = output.max(1)[1].type_as(labels)
    correct = preds.eq(labels).double()
    correct = correct.sum()
    return correct / len(labels)

def accuracy_near(predlabels, labels):
    correct = predlabels.eq(labels).double()
    correct = correct.sum()
    return correct / len(labels)

def triplet_samples(sampleNum, lables):
    ## this function aims at sampling from original data for calculating triplet loss
    ## method: randomly select P samples, for each sample, N samples are sampled.
    ## For each sample within P, N samples are divided into two partitions(positive（A） and negative (B)).
    ## Loss: L_th = 1/(PK) \sigma_{a \in batch} (max_{p \in A} d_{a,p} - min_{n \in B} d_{a,n} + \alpba)_{+}.

    K = P = int(torch.sqrt(sampleNum))
    P_samples_ind = random.sample(range(len(lables)), P)
    P_samples = lables[P_samples_ind]

    PK_samples = np.zeros((P, K))
    PK_boolean = np.zeros((P, K))
    for Pi in range(P): ## for each P sample, sampling K samples
        tmpSamples = lables[random.sample(range(len(lables)), K)]
        PK_samples[Pi, :] = tmpSamples
        PK_boolean[Pi, :] = np.where(tmpSamples == lables[P_samples], 1, 0)[0]
        ## indicate if sample within P and other samples are same class

    return P_samples, PK_samples, PK_boolean

def triplet_sample_batch(train_labels, batchNum, nclass):
    '''
    Divide all data into batches, Then from the all data, batchsize samples are sampled as rate.
    :param train_labels:
    :param batchNum:
    :param nclass : the number of clusters
    :return: triplet samples: num x 3
    '''
    if np.floor(len(train_labels)/batchNum)>nclass:
        batchsize = np.ceil(len(train_labels)/batchNum)
    else:
        batchsize = nclass

    uniqueSamples = torch.unique(train_labels, return_counts=True) ## number
    rateSamples = uniqueSamples[1].to(torch.float)/torch.sum(uniqueSamples[1]) ## normalize rate
    numSamples = torch.ceil(batchsize * rateSamples) ## samples number for each cluster

    batches = torch.Tensor()
    for _ in range(batchNum):
        # batch = torch.Tensor().cuda() ## each batch
        batch = torch.Tensor()

        ## divide batches
        # batch = []
        for labelId in range(len(uniqueSamples[0])):
            label = uniqueSamples[0][labelId] # label name
            num = numSamples[labelId] # the number of samples in a batch

            ## sample nodes within cluster labelId
            clusterIds = torch.where(train_labels == label)[0] ## idx of 'label' in train_labels
            labelIds = torch.multinomial(clusterIds.to(torch.float), int(num.item()))
            batch = torch.cat((batch, clusterIds[labelIds].to(torch.float)), dim=0)

        # batch = list(itertls.chain(batch)) ## chain
        # then construct the triple array

        # oneBatchTripletNodes = torch.Tensor()
        for label in uniqueSamples[0]:
            targetCluster = torch.where(train_labels[batch.to(torch.long)] == label)[0] ## idx of anchors and positive nodes
            anchorAndposit = torch.Tensor(list(itertls.combinations(batch[targetCluster.to(torch.long)], 2)))
            negSamples = batch[torch.where(train_labels[batch.to(torch.long)] != label)[0].to(torch.long)].detach().cpu()
            tripleNodes = list(itertls.product(anchorAndposit, negSamples))
            oneClassTripletNodesList = [[it[0][0], it[0][1], it[1]] for it in tripleNodes]
            oneClassTripletNodes = torch.Tensor(oneClassTripletNodesList)
            # oneClassTripletNodes = torch.cat((torch.from_numpy(np.vstack((tripleNodes[:, 0]))), torch.from_numpy(np.vstack((tripleNodes[:, 1])))), dim=1)
            # oneClassTripletNodes = torch.cat((oneClassTripletNodes, torch.from_numpy(tripleNodes[:,1])), dim=1)
            batches = torch.cat((batches, oneClassTripletNodes), dim=0)
        # batches.append(oneBatchTripletNodes)

#.detach().cpu().numpy()  torch.cat((torch.Tensor(np.vstack(tripleNodes[:,0])).cuda(), torch.Tensor(tripleNodes[:,1])), dim=1)

    return batches


def triplet_sample_random(train_labels, samplesNum):
    '''
    this function aims to sample from the network as trip-array <anchors, positive, negative>
    More detail, given the labels (k x 1) and the number of samples, we select the anchors and positive nodes from
    nodes with same labels and negative nodes from others for each label.

    For instance, given label<L>, we sample the idx of same labels (idx: index for the given labels).
    idx_loc is the index of samples, which is the location of the given idx.
    Get the idx from the idx_loc( those index is under the given labels, the index of labels).

    :param train_labels: labels: k x 1
    :param samplesNum: the number of samples.
    :return:
    '''
    # train_labels: the labels (array): k x 1
    # train_idx is corresponding to train_labels

    # Len = len(train_labels)
    # idx = torch.Tensor(np.arange(Len))
    triple_samples = torch.Tensor([])
    labelset = torch.unique(train_labels)

    for label in labelset:
        idx_label = torch.where(train_labels == label)[0] # label index for current labels but original data
        eachClassNum = int(np.ceil(samplesNum/len(idx_label)))

        idx = torch.floor(torch.rand(size=(1, eachClassNum)) * len(idx_label)).squeeze(0) ## samples idx from idx_label
        idx_anchor_loc = idx.to(torch.long)
        anchors = idx_label[idx_anchor_loc] ##1. get samples idx,

        idx = torch.floor(torch.rand(size=(1, eachClassNum)) * len(idx_label)).squeeze(0)
        idx_pos_loc = idx.to(torch.long)
        posits = idx_label[idx_pos_loc]

        idx_otherlabel = torch.where(train_labels != label)[0]
        idx_other_loc = torch.floor(torch.rand(size=(1, eachClassNum)) * len(idx_otherlabel)).squeeze(0)
        idx_other_loc = idx_other_loc.to(torch.long)
        negats = idx_otherlabel[idx_other_loc]
        # torch.randn(size=())
        # torch.multinomial()
        # anchors = idx_label.multimial(eachClassNum, True) # sampling node id from current labels
        # anchors = anchors[torch.randperm(eachClassNum)]
        # anchors = train_idx[anchors] # obtaining the original node ID
        #
        # posits = idx_label.multimial(eachClassNum, True)
        # posits = posits[torch.randperm(eachClassNum)]
        # posits = train_idx[posits]
        # idx_otherlabel = torch.where(train_labels != label)[0]
        # negats = idx_otherlabel.multinomial(eachClassNum, True)
        # negats = negats[torch.randperm(eachClassNum)]
        # negats = train_idx[negats]

        diffIdx = anchors != posits
        anchors = anchors[diffIdx]
        posits = posits[diffIdx]
        negats = negats[diffIdx]

        oneClassSamp = torch.zeros((len(anchors)), 3)
        oneClassSamp[:,0] = anchors
        oneClassSamp[:,1] = posits
        oneClassSamp[:,2] = negats

        triple_samples = torch.cat((triple_samples, oneClassSamp), dim=0) # 可能有重复。

    return triple_samples



def computeTripletLoss(output, tripletPoints, margin):
    '''
    compute the triplet loss given the embedding of nodes with labels
    :param output: N x out_features
    :param tripletPoints: k x 3 (anchors, positive nodes, negative nodes)
    :param margin: soft-margin for computing triplet loss.
    :return: triplet loss
    '''
    # output = torch.where(output < 0, torch.full_like(output, 0), output)
    '''
    #2019-8-6 17:29:09 等效变换
    '''
    # output = nn.ReLU()(output)
    ## 这里要计算triplet loss， 但是注意，怎么把相关embedding提取出来哦。
    anchorPointsOutput = output[tripletPoints[:,0].to(torch.long), :]
    positPointsOutput = output[tripletPoints[:,1].to(torch.long), :]
    negatPointsOutput = output[tripletPoints[:,2].to(torch.long), :]
    tripletLoss = F.triplet_margin_loss(anchorPointsOutput, positPointsOutput, negatPointsOutput,margin=margin)
    return tripletLoss


def computeCluterCenter(output, train_idx, train_labels):
    '''
    just compute the cluster center for nodes with labels
    :param output: nclass x N x out_features
    :param train_idx: k x 1
    :param train_label: k x 1 corresponding to train_idx.
    :return: embedding of cluster center (nclass x out_features_final)
    '''
    # it should be output[label, idx, :].sum()!!!
    # avergeNodesLabels = torch.zeros(size=(len(output[:, 0, 0]), len(output[0, 0, :]))).cuda() # nclass x dim
    avergeNodesLabels = torch.zeros(size=(len(output[:, 0, 0]), len(output[0, 0, :])))  # nclass x dim

    sumEachCluster = torch.unique(train_labels, return_counts=True)
    labelNodes = output[:, train_idx,:]
    for clusterId in sumEachCluster[0]:
        clusterNodes = train_labels.eq(clusterId)
        avergeNodesLabels[clusterId, :] = torch.sum(labelNodes[clusterId, clusterNodes,:], 0)/sumEachCluster[1][clusterId]
    return avergeNodesLabels

def l2Similarity(a, b):
    '''
    calcuate the l2-similarity between a and b
    :param a: c x N x o
    :param b: c x N x o
    :return: c x N
    '''
    dist = torch.sqrt(torch.sum((a - b) ** 2, axis=2))  # nclass x N
    maxvalue = torch.max(dist, axis=0)
    maxvalue = maxvalue[0].unsqueeze(0).repeat(len(a[:, 0, 0]), 1)
    sim = (maxvalue-dist)/maxvalue # nclass x N
    ## 这里归一化。
    sumvalue = torch.sum(sim, axis=0)
    sumvalue = sumvalue.unsqueeze(0).repeat(len(a[:, 0, 0]), 1)
    sim = sim / sumvalue
    # assert not torch.isnan(sim).any()
    sim[torch.isnan(sim)] = 0.0
    sim[torch.isinf(sim)] = 0.0
    return sim

def nearestEmbedding(output, clusterCenters):
    '''
    For each node, find the nearest node from clusterCenters and label its labels.
    :param output: nclass x N x out_features
    :param clusterCenters: nclass x out_features.
    :return: N x 1
    '''
    N = len(output[0,:,0])
    clusterCenters = clusterCenters.unsqueeze(1).repeat(1, N, 1)
    dist = torch.sqrt(torch.sum((output - clusterCenters) ** 2, axis=2))  # nclass x N
    # sim = l2Similarity(output, clusterCenters) # nclass x N
    return torch.min(dist, 0).indices

def entropy(p, dim = -1):
    ## 计算熵可能有问题！！  维度为-1，可能有问题。 need to check !!!!!!!!!!!!!!!!!
    p = p + 1e-10
    b = p.log()
    # b[torch.isinf(b)] = 0.0
    a = p * b
    # a[torch.isnan(a)] = 0.0
    # a[torch.isinf(a)] = 0.0
    return -torch.where(p > 0, a, p.new([0.0])).sum(dim = dim) # can be a scalar, when PyTorch.supports it

def computeCluterCenter2(output, idx, train_labels):
    '''
    compute the cluster center under all map (only the their corresponding map is aviluable)
    :param output: nclass x N x out_features
    :param idx: k x 1
    :param train_labels: k x 1
    :return: embedding of cluster centers(most of them is fake)
    '''
    nclass = len(output[:, 0, 0])
    # avergeNodesLabels = torch.Tensor().cuda()
    avergeNodesLabels = torch.Tensor()
    # avergeNodesLabels = torch.zeros(size=(nclass * nclass, len(output[0, 0, :]))).cuda()
    sumEachCluster = torch.unique(train_labels, return_counts=True)
    labelNodes = output[:, idx, :]
    for clusterId in sumEachCluster[0]:
        clusterNodes = train_labels.eq(clusterId)
        avetmp = (torch.sum(labelNodes[:, clusterNodes, :], 1)/sumEachCluster[1][clusterId]).unsqueeze(dim=1)
        avergeNodesLabels = torch.cat((avergeNodesLabels, avetmp),dim=1)  # nclass x 1 x out

    return avergeNodesLabels.view(nclass * nclass, len(output[0, 0, :]))

def computeEntropy2(clusterCenters, output, idx=None):
    '''
    Since each node belongs to one cluster based on special map, the embedding of node is similar to
    the cluster center under their corresponding map but the embedding of nodes within other map and
    centers of other cluster.

    reconstruct the (nclass x nclass) x N x out_feature_final
    :param clusterCenters: (nclass x nclass) x out_feature_final
    :param output: nclass x N x out_features_final
    :param idx:
    :return: entropy
    '''
    N = len(output[0, :, 0])
    allIdx = np.linspace(0, N - 1, N).astype(np.int64).tolist()
    if idx is not None:
        idx=idx.tolist()
        restIdx = [int(x) for x in allIdx if int(x) not in idx]
    else: restIdx = [int(x) for x in allIdx]

    clusterCenters = clusterCenters.unsqueeze(1).repeat(1, N, 1)
    output_mul = output.repeat(len(output[:,0,0]), 1, 1)
    sim = l2Similarity(output_mul, clusterCenters)
    entropy_vec = entropy(sim.t(), dim=1)  # N
    loss_entropy = entropy_vec[restIdx].mean()
    return loss_entropy


def computeEntropy(clusterCenter, output, idx=None):
    '''
    compute the entropy between node embeddings and each cluster center
    :param clusterCenter: nclass x out_features_final
    :param output: nclass x N x out_feautures_final
    :param idx: node indicator
    :return:
    '''
    ##MARK for negative figures
    N = len(output[0, :, 0])
    allIdx = np.linspace(0, N - 1, N).astype(np.int64).tolist()
    if idx is not None:
        idx=idx.tolist()
        restIdx = [int(x) for x in allIdx if int(x) not in idx]
    else: restIdx = [int(x) for x in allIdx]

    ##  MARK , for in_place operation
    # consine similarity.
    # cosSim = torch.nn.CosineSimilarity(dim=2, eps=1e-8)
    # clusterCenter = torch.where(clusterCenter < 0, torch.full_like(clusterCenter, 0), clusterCenter)
    # output = torch.where(output < 0, torch.full_like(output, 0), output)

    # l2 distance -> similarity
    clusterCenters = clusterCenter.unsqueeze(1).repeat(1,N,1)

    sim = l2Similarity(output, clusterCenters) ## l2 norm (similarity)
    # cosSim = torch.nn.CosineSimilarity(dim=2, eps=1e-8)
    # sim = cosSim(clusterCenters, output)  # nclass x N x out_features ==> nclass x N
    # sim = sim + 1e-10
    ##MARK  To delete all the negative numbers！！！
    ##sim =torch.where(sim < 0, torch.full_like(sim, 0), sim)
    ##//

    entropy_vec = entropy(sim.t(), dim=1)  # N
    # entropy_vec[torch.isnan(norm_sim)] = 0.0
    # entropy_vec[torch.isinf(norm_sim)] = 0.0
    # loss_entropy = entropy_vec[restIdx].sum()
    loss_entropy = entropy_vec[restIdx].mean()
    # loss_entropy = entropy_vec.sum()
    return loss_entropy

def findNearLabelNodes(output, idx_train, train_labels, k=1, restIdx=None):
    """find the nearest train nodes for each test node. Each node without label find the nearest neighbor in each cluster
    the entropy between given node without label and those nodes should be small, which equals to change the cluster center
    to labelled nodes."""

    if restIdx is None:
        N = len(output[0, :, 0])
        allIdx = np.linspace(0, N - 1, N).astype(np.int64).tolist()
        restIdx = [int(x) for x in allIdx if int(x) not in idx_train]
        restIdx = torch.Tensor(restIdx).to(torch.long).cuda()

    clusterIds = torch.unique(train_labels, return_counts=True)
    # nearLabelNodes = torch.zeros(size=(len(clusterIds[0]), len(restIdx), k)).cuda()  # 存放每个无标签点在每个类中最邻近的k个标签节点ID
    # nearLabelDist = torch.zeros(size=(len(clusterIds[0]), len(restIdx), k)).cuda() # 存放每个无标签节点在每个类中距离最近的点的距离。
    nearLabelNodes = torch.zeros(size=(len(clusterIds[0]), len(restIdx), k))  # 存放每个无标签点在每个类中最邻近的k个标签节点ID
    nearLabelDist = torch.zeros(size=(len(clusterIds[0]), len(restIdx), k)) # 存放每个无标签节点在每个类中距离最近的点的距离。
    for clusterId in clusterIds[0]:
        # for each cluster.
        clusterId_idx_train = idx_train[train_labels.eq(clusterId)]  # index for each labelled node belonged to clusterId
        labelEmbedding = output[clusterId, clusterId_idx_train, :]  # lnum x dim, labelled nodes

        unlabelEmbedding = output[clusterId, restIdx,:] # unum x dim, unlabelled nodes
        lNum = len(labelEmbedding[:,0])
        uNum = len(unlabelEmbedding[:,0])
        l_tensor = labelEmbedding.unsqueeze(dim=0).repeat(uNum, 1, 1) # lnum x dim => unum x lnum x dim
        u_tensor = unlabelEmbedding.unsqueeze(dim=1).repeat(1, lNum, 1) # unum x dim => unum x lnum x dim
        dist = torch.norm((l_tensor-u_tensor), dim=2, p=2) # dist matrix: unum x lnum
        nearNodes = torch.topk(dist, k=k, dim=1, largest=False) # get the smallest k indics: => unum x k
        # the indices and dist for each unlabelled node

        nearIdx = nearNodes[1] # idx of near node Id
        nearDist = nearNodes[0] # dist of near node distance. # unum x k
        nearLabelNodes[clusterId,:,:] = clusterId_idx_train[nearIdx] # the position of the nearest node without label
        nearLabelDist[clusterId,:,:] = nearDist #  distance between given node and its nearest neighbors. nclass x restId x k, k=1

    return nearLabelNodes, nearLabelDist

def stast(nearNodeDist, k):
    """statistics the most frequent elements, if unexisting, return the first element"""
    # this method aims to statistics the elements for each row
    nearLabels = nearNodeDist[1] / k  # N(unlabel) x k_nn (label)
    nearLabels = nearLabels.detach().cpu().numpy()
    nearDist = nearNodeDist[0].detach().cpu().numpy()
    N = nearLabels.shape[0]

    topIdx = np.zeros(N)
    for row in range(N):
        stastCluter = np.bincount(nearLabels[row,:])
        moreIdx = stastCluter == np.max(stastCluter)
        if np.sum(moreIdx) > 1:
            meanDist = []
            for moreLoc in np.where(moreIdx)[0]:
                meanDist.append(np.mean(nearDist[row, nearLabels[row,:] == moreLoc]))
            topIdx[row] = np.where(moreIdx)[0][np.argmin(meanDist)]
        # if np.sum(np.bincount(nearLabels[row, :])>1)>0:
        else:
            topIdx[row] = np.argmax(stastCluter)
        # else:
        #     topIdx[row] = nearLabels[row, 0]

    return topIdx

def constrainCov(X, idx, labels_train, c):
    '''
    covarious, 方差越大，熵越大。
    :param X: (N x dim)
    :param idx: labelled nodes Idx
    :param labels: labelled nodes
    :param c: the number of clusters
    :return:
    '''
    X1 = torch.zeros_like(X[idx,:])
    X2 = torch.zeros_like(X[idx,:])
    # X2 = X2.t()

    X_l = X[idx,:]
    # X2 = X.clone()[idx, :]
    perdim = int(X.shape[1]/c)
    for i in range(c): # dim
        # for j in range(c): # clusters
        #     X1[:, i * perdim: (i + 1) * perdim] -= \
        #         torch.mean(X1[:, i * perdim: (i + 1) * perdim], dim=1).unsqueeze(dim=1).repeat(1, perdim)
            ## all nodes, different features

            X1[labels_train == i, i * perdim: (i + 1) * perdim] = X_l[labels_train == i, i * perdim: (i + 1) * perdim] - \
                torch.mean(X_l[labels_train == i, i * perdim: (i + 1) * perdim], dim=1).unsqueeze(dim=1).repeat(1, perdim)

            X1[labels_train != i, i * perdim: (i + 1) * perdim] = X_l[labels_train != i, i * perdim: (i + 1) * perdim] - \
                torch.mean(X_l[labels_train != i, i * perdim: (i + 1) * perdim], dim=1).unsqueeze(dim=1).repeat(1, perdim)


            # X2[i * perdim: (i + 1) * perdim,labels_train == i ] = X_l.t()[i * perdim: (i + 1) * perdim, labels_train == i] -\
            #     torch.mean(X_l[i * perdim: (i + 1) * perdim, labels_train == i], dim=1).unsqueeze(dim=1).repeat(1, torch.sum(labels_train==i))

            X2[labels_train == i, i * perdim: (i + 1) * perdim] = X_l[labels_train == i, i * perdim: (i + 1) * perdim] -\
                torch.mean(X_l[labels_train == i, i * perdim: (i + 1) * perdim], dim=0).unsqueeze(dim=0).repeat(torch.sum(labels_train==i), 1)

            if i == 0:
                X2[labels_train == i, (i + 1) * perdim :] = X_l[labels_train == 0, (i + 1) * perdim :] - \
                    torch.mean(X_l[labels_train==i, (i + 1) * perdim :], dim=0).unsqueeze(dim=0).repeat(torch.sum(labels_train==i), 1)
            elif i == c-1:
                X2[labels_train == i, :i * perdim] = X_l[labels_train == i,  :i * perdim] - \
                    torch.mean(X_l[labels_train == i,  :i * perdim], dim=0).unsqueeze(dim=0).repeat(torch.sum(labels_train==i), 1)
            else:
                tmp = torch.cat((X_l[labels_train == i,:i * perdim ], X_l[labels_train == i, (i + 1) * perdim :]), dim=1)
                ave = torch.mean(tmp, dim=0).unsqueeze(dim=0)
                X2[labels_train==i, :i * perdim] = X_l[labels_train==i, :i * perdim] - ave[0, :i * perdim].repeat(torch.sum(labels_train == i), 1)
                X2[labels_train == i, (i + 1) * perdim :] = X_l[labels_train == i, (i + 1) * perdim :] - ave[0, i * perdim :].repeat(torch.sum(labels_train == i), 1)

            ## different nodes, all features

    labelsId = torch.unique(labels_train)
    covSum = 0.0
    for clusterId in labelsId:
        inClusters = torch.sum(torch.abs(X1[labels_train == clusterId, clusterId*perdim:(clusterId+1)*perdim] @
                                         X1[labels_train == clusterId, clusterId * perdim:(clusterId + 1) * perdim].t()))
        # in same cluster, cov should be small
        bwClusters = torch.sum(torch.abs(X1[labels_train == clusterId, clusterId*perdim:(clusterId+1)*perdim] @
                               (X1[labels_train != clusterId, clusterId*perdim:(clusterId+1)*perdim]).t()))
        # in different cluster, cov should be large.
        # print(inClusters/perdim)
        covSum += inClusters/bwClusters

        if clusterId == labelsId[0]:
            restFeatures = X2[labels_train == clusterId, (clusterId+1)*perdim :]  ## 除去中间的属性。
        elif clusterId == labelsId[-1]:
            restFeatures = X2[labels_train == clusterId, :clusterId*perdim]
        else:
            restFeatures = torch.cat((X2[labels_train == clusterId, :clusterId*perdim],
                                     X2[labels_train == clusterId, (clusterId+1)*perdim :]), dim=1)

        bwNodes = torch.sum(torch.abs((X2[labels_train == clusterId, clusterId*perdim:(clusterId + 1) * perdim].t()) @ restFeatures))
        inNodes = torch.sum(torch.abs((X2[labels_train == clusterId, clusterId*perdim:(clusterId + 1) * perdim].t()) @
                                      X2[labels_train == clusterId, clusterId * perdim:(clusterId + 1) * perdim]))
        # print(inNodes/torch.sum(labels_train == clusterId))
        covSum += inNodes/bwNodes
    return covSum

# def accuracyEntropy(output, clusterCenter, train_idx, allLabels):
#     '''
#     given output, computing the similarity between the node embedding and other cluster centers.
#     assigning the most similar nodes as the cluster
#     :param output: nclass x N x out_features
#     :param clusterCenter: nclass x out_features
#     :param idx: index of training set
#     :param testLabels:  labels of training set
#     :return: accuray`+
#      of nodes without labels
#     '''
#     clusterCenter = torch.where(clusterCenter < 0, torch.full_like(clusterCenter, 0), clusterCenter)
#     output = torch.where(output < 0, torch.full_like(output, 0), output)
#     N = len(output[0, :, 0])
#     ##MARK for in_place operation
#     ##output[:, train_idx, :] = 0  # remaining node embedding without labels.(nclass x N x out_features_final)
#     cosSim = torch.nn.CosineSimilarity(dim=2, eps=1e-8)
#     # expect to obtain sim matrix (nclass x N)
#     sim = cosSim(clusterCenter.unsqueeze(1).repeat(1, N, 1), output)  # nclass x N x out_features ==> nclass x N
#     value, prelabels = torch.max(sim, dim=0)
#     allIdx = np.linspace(start=0, stop=N - 1, num=N)
#     testIdx = [int(x) for x in allIdx if int(x) not in train_idx]
#
#     prelabels = prelabels[testIdx]  # delete the training set
#     # testLabels = torch.Tensor(allLabels.float())[testIdx].type_as(prelabels)
#     '''
#     #2019-8-7 10:48:50
#     '''
#     testLabels = allLabels[testIdx].type_as(prelabels)
#     correct = testLabels.eq(prelabels).double()
#     return float(correct.sum()) / len(prelabels)

def load_data2(dataset, rate, k):
    """Load data."""
    names = ['x', 'y', 'tx', 'ty', 'allx', 'ally', 'graph']
    objects = []
    for i in range(len(names)):
        with open("data/ind.{}.{}".format(dataset, names[i]), 'rb') as f:
            if sys.version_info > (3, 0):
                objects.append(pkl.load(f, encoding='latin1'))
            else:
                objects.append(pkl.load(f))

    x, y, tx, ty, allx, ally, graph = tuple(objects)
    test_idx_reorder = parse_index_file("data/ind.{}.test.index".format(dataset))  # test ids
    test_idx_range = np.sort(test_idx_reorder)  # change location of test into sort.

    if dataset == 'citeseer':
        # Fix citeseer dataset (there are some isolated nodes in the graph)
        # Find isolated nodes, add them as zero-vecs into the right position
        test_idx_range_full = range(min(test_idx_reorder), max(test_idx_reorder) + 1)
        tx_extended = sp.lil_matrix((len(test_idx_range_full), x.shape[1]))
        tx_extended[test_idx_range - min(test_idx_range), :] = tx
        tx = tx_extended
        ty_extended = np.zeros((len(test_idx_range_full), y.shape[1]))
        ty_extended[test_idx_range - min(test_idx_range), :] = ty
        ty = ty_extended

    features = sp.vstack((allx, tx)).tolil()  ## all features
    features[test_idx_reorder, :] = features[test_idx_range, :]
    adj = nx.adjacency_matrix(nx.from_dict_of_lists(graph))

    labels = np.vstack((ally, ty))
    labels[test_idx_reorder, :] = labels[test_idx_range, :]

    idx_test = test_idx_range.tolist()
    idx_train = range(len(y))
    idx_val = range(len(y), len(y) + 500)

    train_mask = sample_mask(idx_train, labels.shape[0])
    val_mask = sample_mask(idx_val, labels.shape[0])
    test_mask = sample_mask(idx_test, labels.shape[0])

    y_train = np.zeros(labels.shape)
    y_val = np.zeros(labels.shape)
    y_test = np.zeros(labels.shape)
    y_train[train_mask, :] = labels[train_mask, :]
    y_val[val_mask, :] = labels[val_mask, :]
    y_test[test_mask, :] = labels[test_mask, :]

    adj = adj + adj.T.multiply(adj.T > adj) - adj.multiply(adj.T > adj)
    features = normalize(features)

    labels = torch.max(torch.LongTensor(labels), 1)[1]
    idx_train, idx_test, minclusterNum = selectNodes(labels, rate, k)

    return adj, features, idx_train, idx_test, labels, minclusterNum

def selectNodes(labels, rate, k):

    train = []
    clusterId = torch.unique(labels)
    for cid in clusterId:
        numc = np.min((int(torch.sum(labels == cid)), k)) # 如果k比每个类都大。取类个数
        numc = np.max((torch.ceil(torch.sum(labels == cid).to(float) * rate), numc)) # samples number 应该采样的数

        cIdx = torch.where(labels == cid)[0]
        loc = torch.randint(low=0, high=len(cIdx), size=(1, int(numc)))
        train.append(cIdx[loc])
    trainIdx = torch.cat(train, dim=1).squeeze()
    testIdx = [x for x in range(len(labels)) if x not in trainIdx]

    return trainIdx, torch.Tensor(testIdx), numc

def normalize(mx):
    """Row-normalize sparse matrix"""
    rowsum = np.array(mx.sum(1))
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    mx = r_mat_inv.dot(mx)
    return mx